package top.ninwoo.edgecenter.service.impl;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import top.ninwoo.edgecenter.entity.ClusterConfig;
import top.ninwoo.edgecenter.entity.NetworkTopology;
import top.ninwoo.edgecenter.service.ClusterService;
import top.ninwoo.edgecenter.service.IpService;
import top.ninwoo.edgecenter.service.TopologyService;
import top.ninwoo.utils.service.OVSService;
import top.ninwoo.utils.service.OvsDockerService;

import java.util.Set;

/**
 * @Author joliu
 * @Description 这里描述的物理连接的拓扑关系
 * @Date　Create in 上午11:16 2019/11/7
 */
@Service
public class TopologyServiceImpl implements TopologyService {
    private final static Logger LOG = LoggerFactory.getLogger(TopologyServiceImpl.class);

    @Autowired
    OVSService ovsService;

    @Autowired
    OvsDockerService ovsDockerService;

    // 这个服务接口的具体功能应该下沉
    @Autowired
    ClusterService clusterService;

    @Autowired
    IpService ipService;

    @Override
    public void initTopo(ClusterConfig clusterConfig) {
        // 获取topology
        createTopo(clusterConfig.getId(), clusterConfig.getTopology());
    }

    @Override
    public void createTopo(long clusterId, NetworkTopology topology) {
        if(topology == null) {
            LOG.warn("集群[" + clusterId + "]未设置网络");
            return;
        }

        // 校验信息
        String[] cNames = topology.getAppNames();
        int[][] topo = topology.getTopology();
        if(cNames == null || cNames.length <= 1) {
            return;
        }

        // 这里声明：cNames中以br:开头的名称为ovs虚拟交换机
        for (int i = 1; i < cNames.length; i++) {
            for (int j = 0; j < i; j++) {
                // 判断每一个节点的连接状态
                if(topo[i][j] == 1) {
                    // 如果存在连接，则创建连接
                    addLink(clusterId, cNames[i], cNames[j]);
                }
            }
        }
    }

    public void addLink(long clusterId, String appName1, String appName2) {
        int flag = 0;
        // 校验 参数是否合法
        // step1: 判断appNames中是否存在ovs虚拟交换机
        if(appName1.startsWith("br:")) {
            // flag计数器加1
            flag++;
            // 创建对应的虚拟交换机
            // 在创建前需要检查，是否已经创建过对应的交换机
            if(!ovsService.isBridge(appName1)) {
                LOG.info("Create Ovs Bridge!");
                ovsService.addBridge(appName1.substring(3));
            }
            LOG.info("Ovs Bridge[" + appName1 + "] is normal!");
        }

        if(appName2.startsWith("br:")) {
            flag++;
            // 创建对应的虚拟交换机
            if(!ovsService.isBridge(appName2.substring(3))) {
                LOG.info("Create Ovs Bridge!");
                ovsService.addBridge(appName2.substring(3));
            }
            LOG.info("Ovs Bridge[" + appName2 + "] is normal!");
        }

        // 如果两个都是虚拟交换机，则需要创建交换机连接
        if(flag == 2) {
            // 校验是否是远端的交换机
            // 远端交换机输入样例：br:remote:testBr:192.168.31.111
            String remoteIp1 = "";
            if(appName1.contains("remote")) {
                remoteIp1 = appName1.split(":")[3];
            }
            String remoteIp2 = "";
            if(appName2.contains("remote")) {
                remoteIp2 = appName2.split(":")[3];
            }

            // 两个都是本地交换机
            if("".equals(remoteIp1) && "".equals(remoteIp2)) {
                // 创建两个虚拟交换机
                ovsService.addBridge(appName1);
                ovsService.addBridge(appName2);
                // 创建两个交换机之间的连接
                ovsService.linkOvs(appName1, appName2);
            }

            // 如果两个都是远程交换机
            else if(!"".equals(remoteIp1) && !"".equals(remoteIp2)) {
                // do nothing
            }
            // 如果其中之一是本地交换机，找到本地交换机，设置远端ip
            else {
                if("".equals(remoteIp1)) {
                    // appName2是远程的
                    ovsService.setVxlan(appName1, remoteIp2);
                } else if("".equals(remoteIp2)) {
                    // appName1是远程的
                    ovsService.setVxlan(appName2, remoteIp1);
                }
            }

        }
        // 如果其中一个是虚拟交换机，则创建虚拟交换机到容器的连接
        else if(flag == 1) {
            String ovsName = appName1.startsWith("br:") ? appName1.substring(3) : appName2.substring(3);
            String containerName = appName1.startsWith("br:") ? appName2 : appName1;

            // 使用ovsDocker接口创建连接
            // 获取全部的APP容器id
            Set<String> cids = clusterService.getContainerIdsByClusterId(clusterId, containerName);
            cids.forEach( cid -> {
                // ip分配服务
                // TODO: 网段应该是和网络拓扑绑定到一起，需要重新进行设计
                String ip = ipService.assignIpString("10.10.1.0/24", cid);
                // ovsDockerService.addPort(ovsName, "eth1", containerName, "ip");
                ovsDockerService.addPort(ovsName, "eth1", cid, ip);
            });
        }
        // TODO: 如果两个都是容器，这里可以配置逻辑连接
        else {
            throw new RuntimeException("Physical topology does not support c2c link.");
        }
    }
}
