package top.ninwoo.utils.service.impl;

import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import top.ninwoo.utils.entity.ChainEntity;
import top.ninwoo.utils.entity.ChainType;
import top.ninwoo.utils.entity.TableType;
import top.ninwoo.utils.service.IptablesService;
import top.ninwoo.utils.util.IptablesUtils;

import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * @Author joliu
 * @Description 这个服务的实现类，并没有采用定时线程进行更新，
 *  采用了一种类似缓存的实现方案，难点要保持每次操作和Dockeriptables操作的一致性
 *  TODO: 未测试
 * @Date　Create in 上午10:08 2019/10/30
 */
@Service
public class IptablesServiceImpl implements IptablesService {
    // 以容器为单位维护着每个容器的iptables
    private ConcurrentHashMap<String, Map<String, List<ChainEntity>>> allIptables
            = new ConcurrentHashMap<>();
    private final ReentrantReadWriteLock rwLock = new ReentrantReadWriteLock();


    @Autowired
    IptablesUtils iptablesUtils;

    // 这个服务中不需要进行任务的定时更新，默认其他人无法访问docker容器中的配置

    /**
     * 根据容器Id查询全部的iptables信息
     *  这里的容器id，要求必须存在，需要在调用前进行验证
     * @param containerId
     * @return
     */
    @Override
    public Map<String, List<ChainEntity>> getIptablesByContainerId(String containerId) {
        checkCache(containerId);
        try {
            rwLock.readLock().lock();
            return allIptables.get(containerId);
        } finally {
            rwLock.readLock().unlock();
        }
    }

    /**
     * 根据容器Id查询iptables信息
     * @param containerId 容器id
     * @param chainType chain的类型
     * @return
     */
    @Override
    public List<ChainEntity> getIptablesByContainerId(String containerId, ChainType chainType) {
        checkCache(containerId);
        try {
            rwLock.readLock().lock();
            return allIptables.get(containerId).get(chainType.toString());
        } finally {
            rwLock.readLock().unlock();
        }

    }

    public ChainEntity getIptablesByContainerId(String containerId, ChainType chainType, int lineNumber) {
        checkCache(containerId);
        List<ChainEntity> chains = getIptablesByContainerId(containerId, chainType);
        if(chains.size() < lineNumber) {
            return null;
        }
        return chains.get(lineNumber);
    }

    /**
     * 校验缓存是否存在，如果不存在的话，则创建一次查询，并添加到缓存中
     * @param containerId
     */
    @Override
    public void checkCache(String containerId) {
        if(!allIptables.containsKey(containerId)) {
            // 如果查询不到，证明该容器之前没有被操作过，调用底层命令，进行一次查询
            synchronized (allIptables) {
                if(!allIptables.containsKey(containerId)) {
                    Map<String, List<ChainEntity>> containerIptables = iptablesUtils.showIptablesDetail(containerId);

                    allIptables.put(containerId, containerIptables);
                }
            }
        }
    }

    /**
     * 使用chainentity删除对应的iptable
     * @param containerId
     * @param chainEntity
     */
    @Override
    public void deleteIptable(String containerId, ChainEntity chainEntity) {
        checkCache(containerId);
        try {
            rwLock.writeLock().lock();
            iptablesUtils.delIptable(containerId, chainEntity.getType(), chainEntity.getSource(), chainEntity.getDestination(), chainEntity.getTarget());
            // 更新一下docker的iptables现状
            deleteIptablesFromList(containerId, chainEntity);
        } finally {
            rwLock.writeLock().unlock();
        }
    }

    /**
     * 更新iptables容器,有安全问题，如果有其他人插入，这边删除是有安全问题的 TODO:
     * @param containerId
     * @param chainEntity
     */
    @Override
    public void deleteIptablesFromList(String containerId, ChainEntity chainEntity) {
        List<ChainEntity> chainEntities = allIptables.get(containerId).get(chainEntity.getType().toString());
        for (int i = chainEntity.getId(); i < chainEntities.size(); i++) {
            chainEntities.get(i).setId(chainEntities.get(i).getId() - 1);
        }
        chainEntities.remove(chainEntity.getId() - 1);
    }

    /**
     * 通过行号进行删除
     * @param containerId
     * @param type
     * @param lineNumber
     */
    @Override
    public void deleteIptableById(String containerId, ChainType type, int lineNumber) {
        checkCache(containerId);
        // 删除第lineNumber个条目
        try {
            rwLock.writeLock().lock();
            iptablesUtils.delIptable(containerId, TableType.filter, type, lineNumber);
            allIptables.get(containerId).get(type.toString()).remove(lineNumber - 1);
        } finally {
            rwLock.writeLock().unlock();
        }
    }


    /**
     * 插入一个iptable, 插到前边
     * @param containerId
     * @param chainEntity
     * @return
     */
    public ChainEntity insertIptable(String containerId, ChainEntity chainEntity) {
        checkCache(containerId);

        try {
            rwLock.writeLock().lock();
            // 先写入Docker
            iptablesUtils.addIptable(containerId, "insert", chainEntity.getType(),
                    chainEntity.getSource(), chainEntity.getDestination(), chainEntity.getTarget());
            // 再写入all iptables
            chainEntity.setId(1);
            insertIntoIptables(containerId, chainEntity);
        } finally {
            rwLock.writeLock().unlock();
        }
        return chainEntity;
    }

    /**
     * 插入到iptables
     * @param containerId
     * @param chainEntity
     */
    private void insertIntoIptables(String containerId, ChainEntity chainEntity) {
        List<ChainEntity> chainEntities = allIptables.get(containerId).get(chainEntity.getType());
        chainEntities.forEach(c -> {c.setId(c.getId() + 1);});
        chainEntities.add(0, chainEntity);
    }

    /**
     * 在后边追加一个iptable，插到后边
     * @param containerId
     * @param chainEntity
     * @return
     */
    public ChainEntity appendIptable(String containerId, ChainEntity chainEntity) {
        checkCache(containerId);

        try {
            rwLock.writeLock().lock();
            // 先写入Docker
            iptablesUtils.addIptable(containerId, "append", chainEntity.getType(),
                    chainEntity.getSource(), chainEntity.getDestination(), chainEntity.getTarget());
            // 再添加alliptables最后一个
            List<ChainEntity> chainEntities = allIptables.get(containerId).get(chainEntity.getType());
            chainEntity.setId(chainEntities.size());
            chainEntities.add(chainEntity);
        } finally {
            rwLock.writeLock().unlock();
        }
        return chainEntity;
    }
}
